import datetime
import duckdb
from typing import NamedTuple
from pathlib import Path
from . import interpreter, PATH

REQUIRED_FIELDS = ['org_id', 'env_id', 'branch_id']

class Source(NamedTuple):
    src_type: str
    connect_sql: str
    init_schema: bool

DEFAULT_SOURCES = [
    Source('core', "ATTACH ':memory:' AS src;", True),
]


class Client():
    branch_source_exists: bool = False
    events_source_exists: bool = False
    # Manually set to point local events at the events source
    events_source_local: bool = False

    def __init__(
            self,
            config: dict,
            sources: list[Source] | None = DEFAULT_SOURCES,
            con: duckdb.DuckDBPyConnection | None = None,
        ):

        if (con == None):
            self.con = duckdb.connect()
        else:
            self.con = con

        # TODO Would be better typed
        self._config = {
            'org_id': None,
            'env_id': None,
            'branch_id': None,
            'system_dt': (datetime.datetime.now() + datetime.timedelta(days=30)).isoformat(),
            'rating_as_of_dt': datetime.datetime.now().isoformat(),
            'rating_range': [],
            'contract_ids': [],
            'customer_ids': [],
            'product_uids': [],
            'dedup_branch_events': False
        }
        self.set_config(config)
        
        for source in sources:
            self.set_source(source)

        # Validate required fields
        missing_fields = {field: self._config[field] for field in REQUIRED_FIELDS if self._config[field] is None}
        if missing_fields:
            raise Exception(f'Missing one of the required configuration fields: {missing_fields}')

        self.interpreter = interpreter.Interpreter()

    @property
    def config(self):
        return self._config.copy()
    
    def set_source(self, source: Source):
        src_type, connect_sql, init_schema = source

        if src_type == 'core':
            self.con.execute('USE memory; DETACH DATABASE IF EXISTS src;')

            if connect_sql:
                self.con.execute(connect_sql)
            else:
                self.con.execute("ATTACH ':memory:' AS src;")

            if init_schema:
                schema = Path(f'{PATH}/bootstrap_sql/0_init.sql').read_text()
                self.con.execute('USE src;')
                self.con.execute(schema)
        elif src_type == 'branch':
            self.con.execute('DETACH DATABASE IF EXISTS brch;')

            # The interpreter will use the src database in the case there is no brch
            if connect_sql:
                self.con.execute(connect_sql)

                if init_schema:
                    schema = Path(f'{PATH}/bootstrap_sql/0_init.sql').read_text()
                    self.con.execute('USE brch;')
                    self.con.execute(schema)
                    self.con.execute('USE memory;')
                
                self.branch_source_exists = True
            else:
                self.branch_source_exists = False
        elif src_type == 'events':
            self.con.execute('USE memory; DETACH DATABASE IF EXISTS evt;')

            # The interpreter will use the src database in the case there is no evt
            if connect_sql:
                self.con.execute(connect_sql)
                self.events_source_exists = True
            else:
                self.events_source_exists = False
    
    def set_config(self, config_updates: dict):
        # Handle unknown fields
        for key, _ in config_updates.items():
            if key not in self._config:
                raise Exception(f'Unknown fields can not be set: {key}')
        
        # Update configuration with new values, maintaining defaults
        for key, _ in self._config.items():
            new_value = config_updates.get(key)
            if key in config_updates:
                if key in REQUIRED_FIELDS and new_value == None:
                    raise Exception(f'Required field can not be set to None')

                if key in ('branch_source_connect', 'core_source_connect'):
                    self.con.execute(new_value)

                self._config[key] = new_value

    def execute(self, query):
        variables = self.config
        variables['branch_source_exists'] = self.branch_source_exists
        variables['events_source_exists'] = self.events_source_exists
        variables['events_source_local'] = self.events_source_local

        resolved_query = self.interpreter.exec(variables, query)
        return self.con.execute(resolved_query)

    def get_price_span_date_range(self, product_types: tuple): 
        # Takes a list of product types to include (EVENT, FIXED)
        return self.execute(f"""
            SELECT date_trunc('month', MIN(started_at))::timestamp, date_trunc('month', MAX(ended_at))::timestamp
            FROM bframe.price_spans
            WHERE product_type IN {str(product_types)}
        """).fetchone()




import json
import re
import sqlparse
from . import PATH
from pathlib import Path
from jinja2 import Template


def format_array(array: list[str], is_str: bool):
    if (array):
        if (is_str):
            str_items = "', '".join(array)
            return f"['{str_items}']"
        else:
            str_items = ", ".join(array)
            return f"[{str_items}]"
    
    return '[]'


class Interpreter:
    def add_table_template(self, name, template):
        """Add a single template to the dictionary
        
        Args:
            name (str): Name of the template
            template (str): Either a file path or SQL string
        """
        if not isinstance(name, str) or not isinstance(template, str):
            raise TypeError("Name and template must be strings")
        
        self._table_templates[name] = template

    _table_templates = {
        '_raw_products': Path(f'{PATH}/client_sql/_raw_products.sql').read_text(),
        '_local_products': Path(f'{PATH}/client_sql/_local_products.sql').read_text(),
        'products': Path(f'{PATH}/client_sql/products.sql').read_text(),
        '_raw_customers': Path(f'{PATH}/client_sql/_raw_customers.sql').read_text(),
        '_local_customers': Path(f'{PATH}/client_sql/_local_customers.sql').read_text(),
        'customers': Path(f'{PATH}/client_sql/customers.sql').read_text(),
        '_raw_pricebooks': Path(f'{PATH}/client_sql/_raw_pricebooks.sql').read_text(),
        '_local_pricebooks': Path(f'{PATH}/client_sql/_local_pricebooks.sql').read_text(),
        'pricebooks': Path(f'{PATH}/client_sql/pricebooks.sql').read_text(),
        '_raw_list_prices': Path(f'{PATH}/client_sql/_raw_list_prices.sql').read_text(),
        '_local_list_prices': Path(f'{PATH}/client_sql/_local_list_prices.sql').read_text(),
        'list_prices': Path(f'{PATH}/client_sql/list_prices.sql').read_text(),
        '_raw_contracts': Path(f'{PATH}/client_sql/_raw_contracts.sql').read_text(),
        '_local_contracts': Path(f'{PATH}/client_sql/_local_contracts.sql').read_text(),
        'contracts': Path(f'{PATH}/client_sql/contracts.sql').read_text(),
        '_raw_contract_prices': Path(f'{PATH}/client_sql/_raw_contract_prices.sql').read_text(),
        '_local_contract_prices': Path(f'{PATH}/client_sql/_local_contract_prices.sql').read_text(),
        'contract_prices': Path(f'{PATH}/client_sql/contract_prices.sql').read_text(),
        '_raw_events': Path(f'{PATH}/client_sql/_raw_events.sql').read_text(),
        '_local_events': Path(f'{PATH}/client_sql/_local_events.sql').read_text(),
        'events': Path(f'{PATH}/client_sql/events.sql').read_text(),
        'prices': Path(f'{PATH}/client_sql/prices.sql').read_text(),
        'processed_events': Path(f'{PATH}/client_sql/processed_events.sql').read_text(),
        'matched_events': Path(f'{PATH}/client_sql/matched_events.sql').read_text(),
        '_all_price_spans': Path(f'{PATH}/client_sql/_all_price_spans.sql').read_text(),
        'price_spans': Path(f'{PATH}/client_sql/price_spans.sql').read_text(),
        'rated_events': Path(f'{PATH}/client_sql/rated_events.sql').read_text(),
        'event_line_items': Path(f'{PATH}/client_sql/event_line_items.sql').read_text(),
        'fixed_line_items': Path(f'{PATH}/client_sql/fixed_line_items.sql').read_text(),
        'line_items': Path(f'{PATH}/client_sql/line_items.sql').read_text(),
        'invoices': Path(f'{PATH}/client_sql/invoices.sql').read_text(),
        'branches': Path(f'{PATH}/client_sql/branches.sql').read_text(),
        'environments': Path(f'{PATH}/client_sql/environments.sql').read_text(),
        'organizations': Path(f'{PATH}/client_sql/organizations.sql').read_text(),
        '_product_filters': Path(f'{PATH}/client_sql/_product_filters.sql').read_text(),
    }

    # Remove comments
    def comment_replacement(self, query: str):
        # TODO we need to handle comments inside of the views
        # Also comments at the end of the file are fully breaking (need a return at the end of the file)
        return sqlparse.format(query, strip_comments=True).strip()

    def var_replacement(self, query: str, vars: dict):
        full_match = re.search(r"_BF_(\w*\b)", query)

        # No matches we return the query back
        if (full_match is None):
            return query

        new_query = query
        variable_name = full_match.group(1)
        match variable_name:
            case 'BRANCH_ID':
                new_query = re.sub(full_match.group(), str(vars.get('branch_id')), new_query)
            case 'ORG_ID':
                new_query = re.sub(full_match.group(), str(vars.get('org_id')), new_query)
            case 'ENV_ID':
                new_query = re.sub(full_match.group(), str(vars.get('env_id')), new_query)
            case 'SYSTEM_DT':
                new_query = re.sub(full_match.group(), f"'{vars.get('system_dt')}'", new_query)
            case 'RATING_AS_OF_DT':
                new_query = re.sub(full_match.group(), f"'{vars.get('rating_as_of_dt')}'", new_query)
            case 'DEDUP_BRANCH_EVENTS':
                new_query = re.sub(full_match.group(), str(vars.get('dedup_branch_events')), new_query)
            case 'RATING_RANGE_START':
                rating_range_start = "''"
                if (len(vars.get('rating_range')) == 2):
                    rating_range_start = f"'{vars.get('rating_range')[0]}'"
                new_query = re.sub(full_match.group(), rating_range_start, new_query)
            case 'RATING_RANGE_END':
                rating_range_end = "''"
                if (len(vars.get('rating_range')) == 2):
                    rating_range_end = f"'{vars.get('rating_range')[1]}'"
                new_query = re.sub(full_match.group(), rating_range_end, new_query)
            case 'CONTRACT_IDS':
                new_query = re.sub(full_match.group(), format_array(vars.get('contract_ids'), True), new_query)
            case 'CUSTOMER_IDS':
                new_query = re.sub(full_match.group(), format_array(vars.get('customer_ids'), True), new_query)
            case 'PRODUCT_UIDS':
                new_query = re.sub(full_match.group(), format_array(vars.get('product_uids'), False), new_query)
            case 'BRANCH_SOURCE_EXIST':
                exists = 'false'
                if vars.get('branch_source_exists'):
                    exists = 'true'
                new_query = re.sub(full_match.group(), exists, new_query)
            case 'EVENTS_SOURCE_EXIST':
                exists = 'false'
                if vars.get('events_source_exists'):
                    exists = 'true'
                new_query = re.sub(full_match.group(), exists, new_query)
            case 'EVENTS_SOURCE_LOCAL':
                exists = 'false'
                if vars.get('events_source_local'):
                    exists = 'true'
                new_query = re.sub(full_match.group(), exists, new_query)
            case _:
                raise Exception(f"Unknown variable: {variable_name}")
        
        return self.var_replacement(new_query, vars)


    def table_replacement(self, query, count=0, verbose=False):
        if (verbose):
            full_match = re.search(r"(\n( *).*)?(bframe\.(\w*\b))", query)
            if (full_match is None):
                return query
            spaces = full_match.group(2) or ""
            schema_table = full_match.group(3)
            table = full_match.group(4)
        else:
            full_match = re.search(r"(bframe\.(\w*\b))", query)
            if (full_match is None):
                return query
            schema_table = full_match.group(1)
            table = full_match.group(2)

        # No matches we return the query back

        if table not in self._table_templates:
            raise Exception(f'An invalid table was used: {table}')
            
        table_template = self._table_templates[table]
        
        if (verbose):
            # Whenever something is placed we must tab as well to add the spaces that already existed
            white_space = "    " + spaces
            table_template = f"(\n{white_space}" + re.sub(r"\n", "\n" + white_space, table_template) + f"\n{spaces})"
        else:
            table_template = f"({table_template})"

        resolved_query = re.sub(schema_table, table_template, query, count=1)

        return self.table_replacement(resolved_query, count=(count+1), verbose=verbose)
    
    def exec(self, vars: dict, query: str, verbose=False):
        resolved_query = self.comment_replacement(query)
        resolved_query = self.table_replacement(resolved_query, verbose=verbose)
        resolved_query = self.var_replacement(resolved_query, vars)
        resolved_query = Template(resolved_query).render()
        
        if (verbose):
            print(resolved_query)

        return resolved_query




import os
import pathlib

PATH = os.path.dirname(__file__)
__version__ = "0.1.4"

# PATH must be exported first since subsequent modules reference it
from .client import Client, Source, DEFAULT_SOURCES
from .interpreter import Interpreter

SCHEMA_SQL = pathlib.Path(f'{PATH}/bootstrap_sql/0_init.sql').read_text()


